(* Convert image (bmp, png, gif, jpg, tif) to Windows icon

    Dr. J. Rathlev, D-24222 Schwentinental

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   J. Rathlev, Nov. 2010
   last modified: March 2018
   *)

unit IconMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Buttons,
  Vcl.ExtCtrls, Vcl.ExtDlgs, Vcl.ImgList;

const
  ProgName = 'Convert image to icon';
  Vers = '3.1';
  CopRgt = ' 2010-2020 Dr. J. Rathlev, D-24222 Schwentinental';
  EmailAdr = 'kontakt(a)rathlev-home.de';

  ImageFormats = '0.jpg;0.jpeg;1.gif;2.png;3.bmp;4.tif;4:.tiff';
  IcoExt = 'ico';

type
  THauptForm = class(TForm)
    Label1: TLabel;
    btnFile: TSpeedButton;
    paImage: TPanel;
    PaintBox: TPaintBox;
    btConvert: TBitBtn;
    OpenPictureDialog: TOpenPictureDialog;
    rgColor: TRadioGroup;
    gbTrans: TGroupBox;
    pnColor: TPanel;
    rbAuto: TRadioButton;
    rbColor: TRadioButton;
    rbOff: TRadioButton;
    SavePictureDialog: TSavePictureDialog;
    btQuit: TBitBtn;
    Label2: TLabel;
    laSize: TLabel;
    rgSize: TRadioGroup;
    btInfo: TBitBtn;
    cbGrid: TCheckBox;
    paRight: TPanel;
    cbImage: TComboBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnFileClick(Sender: TObject);
    procedure PaintBoxPaint(Sender: TObject);
    procedure cbImageCloseUp(Sender: TObject);
    procedure btConvertClick(Sender: TObject);
    procedure rgColorClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure rbTransClick(Sender: TObject);
    procedure PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btQuitClick(Sender: TObject);
    procedure btInfoClick(Sender: TObject);
    procedure rgSizeClick(Sender: TObject);
    procedure cbGridClick(Sender: TObject);
  private
    { Private-Deklarationen }
    FImage        : TBitMap;
    ProgVersName,
    ProgVersDate,
    AppPath,UserPath,
    IniName,ProgPath,
    DestPath,
    LastPath      : string;
    function LoadGrafikFromFile (FName : string) : boolean;
    procedure SetTransparency;
  public
    { Public-Deklarationen }
  end;

var
  HauptForm: THauptForm;

implementation

{$R *.dfm}

uses Vcl.Imaging.jpeg, Vcl.Imaging.GifImg, Vcl.Imaging.PngImage, System.IniFiles,
  WinUtils, TiffImage, IconsToFile, FileUtils, LangUtils, InitProg, GnuGetText;

{ ---------------------------------------------------------------- }
const
  IniExt = 'ini';

  (* INI-Sektionen *)
  CfgSekt  = 'Config';
  DirSekt  = 'Directories';

  (* INI-Variablen *)
  iniLast     = 'LastDir';
  iniDest     = 'DestDir';
  iniTop      = 'Top';
  iniLeft     = 'Left';
  iniHeight   = 'Height';
  iniWidth    = 'Width';
  iniGrid     = 'Grid';
  iniColDpt   = 'ColorDepth';

procedure THauptForm.FormCreate(Sender: TObject);
begin
  TranslateComponent(self);
  Application.Title:=_('Convert image to icon');
  InitPaths(AppPath,UserPath,ProgPath);
  InitVersion(Application.Title,Vers,CopRgt,3,3,ProgVersName,ProgVersDate);
  Caption:=Application.Title+' ('+VersInfo.Comments+')';
  IniName:=Erweiter(AppPath,PrgName,IniExt);
  with TIniFile.Create(IniName) do begin
    LastPath:=ReadString(CfgSekt,IniLast,UserPath);
    DestPath:=ReadString(CfgSekt,IniDest,LastPath);
    Top:=ReadInteger(CfgSekt,iniTop,Top);
    Left:=ReadInteger(CfgSekt,iniLeft,Left);
    ClientHeight:=ReadInteger(CfgSekt,iniHeight,ClientHeight);
    ClientWidth:=ReadInteger(CfgSekt,iniWidth,ClientWidth);
    cbGrid.Checked:=ReadBool(CfgSekt,iniGrid,false);
    rgColor.ItemIndex:=ReadInteger(CfgSekt,iniColDpt,1);
    Free;
    end;
  LoadHistory(IniName,DirSekt,cbImage);
  FImage:=TBitMap.Create;
  end;

procedure THauptForm.FormDestroy(Sender: TObject);
begin
  with TIniFile.Create(IniName) do begin
    WriteString(CfgSekt,IniLast,ExtractFilePath(cbImage.Text));
    WriteString(CfgSekt,IniDest,DestPath);
    WriteInteger(CfgSekt,iniTop,Top);
    WriteInteger(CfgSekt,iniLeft,Left);
    WriteInteger(CfgSekt,iniHeight,ClientHeight);
    WriteInteger(CfgSekt,iniWidth,ClientWidth);
    WriteBool(CfgSekt,iniGrid,cbGrid.Checked);
    WriteInteger(CfgSekt,iniColDpt,rgColor.ItemIndex);
    Free;
    end;
  SaveHistory(IniName,DirSekt,true,cbImage);
  FImage.Free;
  end;

procedure THauptForm.FormShow(Sender: TObject);
begin
  if ParamCount>0 then cbImage.Text:=ParamStr(1);
  btConvert.Enabled:=LoadGrafikFromFile(cbImage.Text);
//  rgSizeClick(Sender);
  end;

procedure THauptForm.PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if rbColor.Checked then begin
    pnColor.Color:=PaintBox.Canvas.Pixels[x,y];
    with FImage do begin
      Transparent:=false;
      TransparentMode:=tmFixed;
      TransparentColor:=pnColor.Color;
      Transparent:=true;
      end;
    PaintBox.Invalidate;
    end;
  end;

procedure THauptForm.PaintBoxPaint(Sender: TObject);
var
  i,n,k,w : integer;
begin
  case rgSize.ItemIndex of
  0 : begin
      k:=FImage.Width;
      if (k=0) or (k>32) then k:=32;
      end;
  1 : k:=16;
  2 : k:=24;
  3 : k:=32;
  4 : k:=48;
  else k:=32;
    end;
  with PaintBox,Canvas do begin
    if Width>Height then n:=Height else n:=Width;
    n:=(n div k)*k;
    StretchDraw(Rect(0,0,n,n),FImage);
    if cbGrid.Checked then begin
      with Pen do begin
        Width:=1;
        Color:=clBlack;
        end;
      w:=n div k ; n:=k*w;
      for i:=0 to k do begin
        MoveTo(0,i*w); LineTo(n,i*w);
        MoveTo(i*w,0); LineTo(i*w,n);
        end;
      end;
    end;
  end;

function THauptForm.LoadGrafikFromFile (FName : string) : boolean;
var
  n  : integer;
  bj : TPicture;
  bg : TGifImage;
  bp : TPngImage;
  bt : TTiffImage;
begin
  Result:=false;
  if (length(FName)>0) then begin
    if FileExists(FName) then begin
      n:=AnsiPos(AnsiLowerCase(ExtractFileExt(FName)),ImageFormats);
      n:=ord(ImageFormats[n-1])-48;
      case n of
      0 : begin //jpeg
          bj:=TPicture.Create;
          with bj do begin
            LoadFromFile(FName);
            Result:=not Graphic.Empty;
            end;
          if Result then FImage.Assign(bj.Graphic);
          bj.Free;
          end;
      1 : begin //Gif
          bg:=TGifImage.Create;
          with bg do begin
            LoadFromFile(FName);
            Result:=not Empty;
            end;
          if Result then FImage.Assign(bg);
          bg.Free;
          end;
      2 : begin //Png
          bp:=TPngImage.Create;
          with bp do begin
            LoadFromFile(FName);
            Result:=not Empty;
            end;
          if Result then FImage.Assign(bp);
          bp.Free;
          end;
      3 : begin //Bmp
          with FImage do begin
            LoadFromFile(FName);
            Result:=not Empty;
            end;
          end;
      4 : begin //Tiff
          bt:=TTiffImage.Create;
          with bt do begin
            LoadFromTifFile(FName);
            Result:=not Empty;
            end;
          if Result then FImage.Assign(bt);
          bt.Free;
          end;
        end;
      if not Result then
        ErrorDialog('Error loading image:'+sLineBreak+FName);
      SetTransparency;
      PaintBox.Invalidate;
      with FImage do begin
        laSize.Caption:=IntToStr(Width)+' x '+IntToStr(Height);
//        case PixelFormat of
//        pf1bit, pf4bit : n:=0;
//        pfDevice, pf8bit : n:=1;
//        else n:=2;
//          end;
//        rgColor.ItemIndex:=n;
        end;
        rgSize.ItemIndex:=0;
      end
    else begin
      ErrorDialog('Image not found:'+sLineBreak+FName);
      cbImage.SetFocus;
      end;
    end;
  end;

procedure THauptForm.SetTransparency;
begin
  PaintBox.Cursor:=crDefault;
  with FImage do if rbAuto.Checked then begin
    Transparent:=true;
    TransparentMode:=tmAuto;
    pnColor.Color:=TransparentColor;
    end
  else if rbColor.Checked then begin
    Transparent:=false;
    TransparentMode:=tmFixed;
    TransparentColor:=pnColor.Color;
    PaintBox.Cursor:=crCross;
    end
  else Transparent:=false;
  end;

procedure THauptForm.rbTransClick(Sender: TObject);
begin
  SetTransparency;
  PaintBox.Invalidate;
  end;

procedure THauptForm.rgColorClick(Sender: TObject);
begin
//  btConvert.Enabled:=LoadGrafikFromFile(hcImage.Text);
  end;

procedure THauptForm.rgSizeClick(Sender: TObject);
begin
  PaintBox.Invalidate;
  end;

procedure THauptForm.btInfoClick(Sender: TObject);
begin
  InfoDialog(ProgName,ProgVersName+' - '+ProgVersDate+sLineBreak+CopRgt);
  end;

procedure THauptForm.btnFileClick(Sender: TObject);
begin
  with OpenPictureDialog do begin
    with cbImage do if length(Text)>0 then
      InitialDir:=ExtractFilePath(Text)
    else if length(LastPath)>0 then InitialDir:=LastPath
    else InitialDir:='';
    FileName:='';
    if Execute then begin
      cbImage.Text:=Filename;
      AddToHistory(cbImage,Filename);
      btConvert.Enabled:=LoadGrafikFromFile(Filename);
      end;
    end;
  end;

procedure THauptForm.btQuitClick(Sender: TObject);
begin
  Close;
  end;

procedure THauptForm.cbGridClick(Sender: TObject);
begin
  PaintBox.Invalidate;
  end;

procedure THauptForm.cbImageCloseUp(Sender: TObject);
var
  s  : string;
begin
  with cbImage do s:=Items[itemindex];
  if (length(s)>0) then begin
    btConvert.Enabled:=LoadGrafikFromFile(s)
    end;
  end;

procedure THauptForm.btConvertClick(Sender: TObject);
var
  ConvIcon : TIcon;
  Bits     : TBitCount;
  bm       : TBitmap;

  function SameSize (n : integer) : boolean;
  begin
    with FImage do Result:=(n=Height) and (n=Width);
    end;

  procedure CopyImage (ASize : integer);
  begin
    with bm do begin
      if SameSize(ASize) or (ASize=0) then Assign(FImage)
      else begin
        Width:=ASize; Height:=ASize;
        Canvas.StretchDraw(Rect(0,0,ASize,ASize),FImage);
        end;
      end;
    end;

begin
  with SavePictureDialog do begin
    if length(DestPath)=0 then InitialDir:=ExtractFilePath(cbImage.Text)
    else InitialDir:=DestPath;
    FileName:=ExtractFilename(NewExt(cbImage.Text,IcoExt));
    if Execute then begin
      DestPath:=ExtractFilePath(Filename);
      bm:=TBitmap.Create;
      case rgColor.ItemIndex of
      0 : Bits:=BitC4;
      2 : Bits:=BitC24;
      else Bits:=BitC8;
        end;
      with bm do begin
        if rbAuto.Checked then TransparentMode:=tmAuto
        else if rbColor.Checked then begin
          TransparentMode:=tmFixed;
          TransparentColor:=FImage.TransparentColor;
          end;
        end;
        case rgSize.Itemindex of
        0 : CopyImage(0);
        1 : CopyImage(16);
        2 : CopyImage(24);
        3 : CopyImage(32);
        4 : CopyImage(48);
        5 : CopyImage(64);
        6 : CopyImage(128);
        7 : CopyImage(256);
        end;
      bm.Transparent:=not rbOff.Checked;
      ConvIcon:=CreateIconFromBitmap(bm);
      if not ConvIcon.Empty then SaveIconToFile(Filename,ConvIcon.Handle,Bits);
      bm.Free;
      ConvIcon.Free;
      end;
    end;
  end;

end.
